/*
 * Copyright (c) 2010 Martin Aigner, Andreas Haas
 * http://cs.uni-salzburg.at/~maigner
 * http://cs.uni-salzburg.at/~ahaas
 *
 * University Salzburg, www.uni-salzburg.at
 * Department of Computer Science, cs.uni-salzburg.at
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include "stm.h"
#include "scm-desc.h"
#include "meter.h"
#include "arch.h"
#include <errno.h>

#ifdef SCM_PRINTMEM
#include <malloc.h>
#endif //SCM_PRINTMEM

#ifdef REGIONALLOC
#include "regmalloc.h"
#endif

static long global_time = 0;
static unsigned int number_of_threads = 0;

//the number of threads, that have not yet ticked in a global period
static unsigned int ticked_threads_countdown = 1;

//protects global_time, number_of_threads and ticked_threads_countdown
static pthread_spinlock_t global_time_lock;

static descriptor_root_t *terminated_descriptor_roots = NULL;

//protects the data structures of terminated threads
static pthread_mutex_t terminated_descriptor_roots_mutex =
        PTHREAD_MUTEX_INITIALIZER;

//the first thread joining libscm has to initialize the thread local storage
static unsigned int init_threads = 0;
static pthread_key_t descriptor_root_key = -1;
static pthread_key_t descriptor_page_pool_key = -1;

static inline void lock_global_time();
static inline void unlock_global_time();
static inline void lock_descriptor_roots();
static inline void unlock_descriptor_roots();
static descriptor_root_t *new_descriptor_root();
static descriptor_root_t *scm_register_thread();
static stm_clock_t *scm_register_thread_clock(pthread_key_t* clock_key);
static descriptor_page_pool_t *scm_register_page_pool();
// TODO: change back: static stm_clock_t* get_clock(pthread_key_t* clock_key);

void *__wrap_malloc(size_t size);
void *__wrap_calloc(size_t nelem, size_t elsize);
void *__wrap_realloc(void *ptr, size_t size);
void __wrap_free(void *ptr);
size_t __wrap_malloc_usable_size(void *ptr);

//avoid ELF interposition of exported but internally used symbols
//by creating static aliases
static void scm_resume_thread_internal(void)
__attribute__((alias("scm_resume_thread")));

static void scm_block_thread_internal(void)
__attribute__((alias("scm_block_thread")));

static void* __wrap_malloc_internal(size_t size)
__attribute__((alias("__wrap_malloc")));

void* scm_malloc(size_t size)
__attribute__((alias("__wrap_malloc")));

void scm_free(void *ptr)
__attribute__((alias("__wrap_free")));

void scm_collect_internal(expired_descriptor_page_list_t *list)
__attribute__((alias("scm_collect")));

void *__wrap_calloc(size_t nelem, size_t elsize) {

    void *p = __wrap_malloc_internal(nelem * elsize);
    //calloc returns zeroed memory by specification
    memset(p, '\0', nelem * elsize);
    return p;
}

/**
 * TODO: change to region-based allocator
 */
void *scm_malloc_clock(size_t size, pthread_key_t* clock_key) {
#ifdef REGIONALLOC
	stm_clock_t* clock = get_clock(clock_key);

	void* region = SCM_REGION_OFFSET(clock->region);
    object_header_t *object =
            (object_header_t*) (regmalloc(size + sizeof (object_header_t), region));

    object->region = clock->region;

    if (!object) {
    	errno = ECANCELED;
        perror("region malloc");
        return NULL;
    }
#else // !REGIONALLOC
    object_header_t *object =
    		(object_header_t*) (__real_malloc(size + sizeof (object_header_t)));

    object->dc = 0;
    object->finalizer_index = -1;
#endif // !REGIONALLOC
    // refreshed with 0 to the clock that has been provided as an argument
    scm_refresh_clock(PAYLOAD_OFFSET(object), 0, clock_key);
#ifdef SCM_PRINTOVERHEAD
    inc_overhead(sizeof (object_header_t));
#endif

#ifdef SCM_PRINTMEM
    printf("[malloc]\n");
    print_memory_consumption();
#endif
    return PAYLOAD_OFFSET(object);
}

void *__wrap_malloc(size_t size) {

    object_header_t *object =
            (object_header_t*) (__real_malloc(size + sizeof (object_header_t)));

    if (!object) {
    	errno = ECANCELED;
        perror("malloc");
        return NULL;
    }
#ifndef REGIONALLOC
    object->dc = 0;
    object->finalizer_index = -1;
#endif

#ifdef SCM_PRINTOVERHEAD
    inc_overhead(sizeof (object_header_t));
#endif

#ifdef SCM_PRINTMEM
    perror("CAUTION: wrapped malloc has been called");
    inc_used_mem(__real_malloc_usable_size(object));
    print_memory_consumption();
#endif
    return PAYLOAD_OFFSET(object);
}

void *__wrap_realloc(void *ptr, size_t size) {

    if (ptr == NULL) return __wrap_malloc_internal(size);
    //else: create new object
    object_header_t *new_object =
            (object_header_t*) __real_malloc(size + sizeof (object_header_t));

    if (!new_object) {
        perror("realloc");
        return NULL;
    }
#ifndef REGIONALLOC
    new_object->dc = 0;
    new_object->finalizer_index = -1;
#endif

#ifdef SCM_PRINTOVERHEAD
    inc_overhead(sizeof (object_header_t));
#endif

    //get the minimum of the old size and the new size
    size_t old_object_size =
            __real_malloc_usable_size(OBJECT_HEADER(ptr))
            - sizeof (object_header_t);
    size_t lesser_object_size;
    if (old_object_size >= size) {
        lesser_object_size = size;
    } else {
        lesser_object_size = old_object_size;
    }

    object_header_t *old_object = OBJECT_HEADER(ptr);
    //copy payload bytes 0..(lesser_size-1) from the old object to the new one
    memcpy(PAYLOAD_OFFSET(new_object),
            PAYLOAD_OFFSET(old_object),
            lesser_object_size);

#ifndef REGIONALLOC
    if (old_object->dc == 0) {
        //if the old object has no descriptors, we can free it
#ifdef SCM_PRINTMEM
        inc_freed_mem(__real_malloc_usable_size(old_object));
#endif

#ifdef SCM_PRINTOVERHEAD
        dec_overhead(sizeof (object_header_t));
#endif
        __real_free(old_object);
    } //else: the old object will be freed later due to expiration
#endif

#ifdef SCM_PRINTMEM
    perror("CAUTION: wrapped realloc has been called");
    inc_used_mem(__real_malloc_usable_size(new_object));
    print_memory_consumption();
#endif
    return PAYLOAD_OFFSET(new_object);
}

void __wrap_free(void *ptr) {

    if (ptr == NULL) return;

    object_header_t *object = OBJECT_HEADER(ptr);
#ifdef REGIONALLOC
    if(object->region == NULL) {
    	__real_free(object);
    }
#else // !REGIONALLOC

    if (object->dc == 0) {
#ifdef SCM_PRINTOVERHEAD
        dec_overhead(sizeof (object_header_t));
#endif
#ifdef SCM_PRINTMEM
        inc_freed_mem(__real_malloc_usable_size(object));
#endif
        __real_free(object);
    }
#endif // !REGIONALLOC
}

size_t __wrap_malloc_usable_size(void *ptr) {
    object_header_t *object = OBJECT_HEADER(ptr);
    return __real_malloc_usable_size(object) - sizeof (object_header_t);
}

/**
 * Initializes a clock. This method is necessary for single-threaded
 * multi-clock mutators.
 * @returns a clock-pointer that identifies the initialized clock
 */
void init_clock(pthread_key_t* clock_key) {
	scm_register_thread_clock(clock_key);
}

void scm_tick_clock(pthread_key_t* clock_key) {
MICROBENCHMARK_START

    descriptor_root_t* descriptor_root = get_descriptor_root_clock(clock_key);

    //make local time progress
    //current_index is equal to the so-called thread-local time
    increment_current_index(&descriptor_root->locally_clocked_buffer);

    //expire_buffer operates on current_index - 1, so it is called after
    //we incremented the current_index of the locally_clocked_buffer
    expire_buffer(&descriptor_root->locally_clocked_buffer,
            &descriptor_root->list_of_expired_descriptors);


#ifndef SCM_EAGER_COLLECTION
    //we also process expired descriptors at tick
    //to get a cyclic allocation/free scheme. this is optional
    expire_descriptor_if_exists(&descriptor_root->list_of_expired_descriptors);
#else
    scm_collect_internal(&descriptor_root->list_of_expired_descriptors);
#endif

#ifdef SCM_PRINTMEM
    printf("[tick]\n");
    print_memory_consumption();
#endif

MICROBENCHMARK_STOP
MICROBENCHMARK_DURATION("scm_tick")
}

void scm_tick(void) {
MICROBENCHMARK_START

    descriptor_root_t *descriptor_root = get_descriptor_root();

    //make local time progress
    //current_index is equal to the so-called thread-local time
    increment_current_index(&descriptor_root->locally_clocked_buffer);

    //expire_buffer operates on current_index - 1, so it is called after
    //we incremented the current_index of the locally_clocked_buffer
    expire_buffer(&descriptor_root->locally_clocked_buffer,
            &descriptor_root->list_of_expired_descriptors);


#ifndef SCM_EAGER_COLLECTION
    //we also process expired descriptors at tick
    //to get a cyclic allocation/free scheme. this is optional
    expire_descriptor_if_exists(&descriptor_root->list_of_expired_descriptors);
#else
    scm_collect_internal(&descriptor_root->list_of_expired_descriptors);
#endif

#ifdef SCM_PRINTMEM
    print_memory_consumption();
#endif

MICROBENCHMARK_STOP
MICROBENCHMARK_DURATION("scm_tick")
}

void scm_global_tick(void) {
MICROBENCHMARK_START
    descriptor_root_t *descriptor_root = get_descriptor_root();

#ifdef SCM_DEBUG
    printf("scm_global_tick GT: %lu GP: %lu #T:%d ttc:%d\n",
            global_time, descriptor_root->global_phase,
            number_of_threads, ticked_threads_countdown);
#endif

    if (global_time == descriptor_root->global_phase) {
        //each thread must expire its own globally clocked buffer,
        //but can only do so on its first tick after the last global
        //time advance

        //my first tick in this global period
        descriptor_root->global_phase++;

        //current_index is equal to the so-called thread-global time
        increment_current_index(&descriptor_root->globally_clocked_buffer);

        //expire_buffer operates on current_index - 1, so it is called after
        //we incremented the current_index of the globally_clocked_buffer
        expire_buffer(&descriptor_root->globally_clocked_buffer,
                &descriptor_root->list_of_expired_descriptors);

        if (atomic_int_dec_and_test((int*) & ticked_threads_countdown)) {

            lock_global_time();

            ticked_threads_countdown = number_of_threads;
            //assert: descriptor_root->global_phase == global_time + 1
            global_time++;

            unlock_global_time();

        } //else global time does not advance

    } //else: we already ticked in this global_phase


#ifndef SCM_EAGER_COLLECTION
    //we also process expired descriptors at tick
    //to get a cyclic allocation/free scheme. this is optional
    expire_descriptor_if_exists(&descriptor_root->list_of_expired_descriptors);
#else
    scm_collect_internal(&descriptor_root->list_of_expired_descriptors);
#endif

#ifdef SCM_PRINTMEM
    print_memory_consumption();
#endif
MICROBENCHMARK_STOP
MICROBENCHMARK_DURATION("scm_global_tick")
}

void scm_collect(expired_descriptor_page_list_t *list) {

    while (expire_descriptor_if_exists(list));

}

static inline int check_extension(int given_extension) {
    if (given_extension > SCM_MAX_EXPIRATION_EXTENSION) {
#ifdef SCM_DEBUG
        printf("violation of SCM_MAX_EXPIRATION_EXTENT\n");
#endif
        return SCM_MAX_EXPIRATION_EXTENSION;
    } else {
        return given_extension;
    }
}

#ifndef REGIONALLOC
void scm_global_refresh(void *ptr, unsigned int extension) {
MICROBENCHMARK_START
#ifdef SCM_DEBUG
    printf("scm_global_refresh(%lx, %d)\n", (unsigned long) ptr, extension);
#endif

    descriptor_root_t *descriptor_root = get_descriptor_root();
    object_header_t *object = OBJECT_HEADER(ptr);

    extension = check_extension(extension);

    atomic_int_inc((int*) & object->dc);
    insert_descriptor(object,
            &descriptor_root->globally_clocked_buffer, extension + 2);

#ifndef SCM_EAGER_COLLECTION
    expire_descriptor_if_exists(&descriptor_root->list_of_expired_descriptors);
#else
    //do nothing. expired descriptors are collected at tick
#endif

#ifdef SCM_PRINTMEM
    print_memory_consumption();
#endif
MICROBENCHMARK_STOP
MICROBENCHMARK_DURATION("scm_global_refresh")
}
#endif

/**
 * TODO: refresh region of region-based allocator with the same extension
 */
void scm_refresh_clock(void *ptr, unsigned int extension, pthread_key_t* clock_key) {
MICROBENCHMARK_START
#ifdef SCM_DEBUG
    printf("scm_refresh(%lx, %d, %d)\n", (unsigned long) ptr, extension, *clock_key);
#endif

descriptor_root_t *descriptor_root = get_descriptor_root_clock(clock_key);

extension = check_extension(extension);
object_header_t* object = OBJECT_HEADER(ptr);

#ifdef REGIONALLOC
	scm_region_header_t* region = object->region;
    atomic_int_inc((int*) & region->dc);
    insert_descriptor_clock(region,
            &descriptor_root->locally_clocked_buffer, extension, clock_key);
#else // !REGIONALLOC
    atomic_int_inc((int*) & object->dc);
    insert_descriptor_clock(object,
            &descriptor_root->locally_clocked_buffer, extension, clock_key);
#endif

#ifndef SCM_EAGER_COLLECTION
    expire_descriptor_if_exists(&descriptor_root->list_of_expired_descriptors);
#else
    //do nothing. expired descriptors are collected at tick
#endif

#ifdef SCM_PRINTMEM
    printf("[refresh]\n");
    print_memory_consumption();
#endif
MICROBENCHMARK_STOP
MICROBENCHMARK_DURATION("scm_refresh_clock")
}

#ifndef REGIONALLOC
void scm_refresh(void *ptr, unsigned int extension) {
MICROBENCHMARK_START
#ifdef SCM_DEBUG
    printf("scm_refresh(%lx, %d)\n", (unsigned long) ptr, extension);
#endif

    descriptor_root_t *descriptor_root = get_descriptor_root();
    object_header_t *object = OBJECT_HEADER(ptr);

    extension = check_extension(extension);

    atomic_int_inc((int*) & object->dc);
    insert_descriptor(object,
            &descriptor_root->locally_clocked_buffer, extension);
    
#ifndef SCM_EAGER_COLLECTION
    expire_descriptor_if_exists(&descriptor_root->list_of_expired_descriptors);
#else
    //do nothing. expired descriptors are collected at tick
#endif

#ifdef SCM_PRINTMEM
    print_memory_consumption();
#endif
MICROBENCHMARK_STOP
MICROBENCHMARK_DURATION("scm_refresh")
}
#endif

// TODO: change back to static
stm_clock_t* get_clock(pthread_key_t* clock_key) {
	stm_clock_t* clock = pthread_getspecific(*clock_key);

	if (clock == NULL) {
		clock = scm_register_thread_clock(clock_key);
	}
	return clock;
}

/*
 * get_descriptor_root is called by operations that need to access the thread
 * local meta data structures.
 */
descriptor_root_t *get_descriptor_root_clock(pthread_key_t* clock_key) {

	stm_clock_t* clock = get_clock(clock_key);
    return clock->descriptor_root;
}

/*
 * get_descriptor_page_pool returns descriptor pages from a thread-local pool
 */
descriptor_page_pool_t *get_descriptor_page_pool() {

    void *ptr = pthread_getspecific(descriptor_page_pool_key);

    if (ptr == NULL) {
        ptr = scm_register_page_pool();
    }

    return (descriptor_page_pool_t*) ptr;
}

/*
 * get_descriptor_root is called by operations that need to access the thread
 * local meta data structures.
 */
descriptor_root_t *get_descriptor_root() {

    void *ptr = pthread_getspecific(descriptor_root_key);

    if (ptr == NULL) {
        ptr = scm_register_thread();
    }

    return (descriptor_root_t*) ptr;
}

/*
 * scm_register_page_pool is called on a thread when the page_pool is needed for the first time
 */
static descriptor_page_pool_t *scm_register_page_pool() {
	descriptor_page_pool_t *page_pool;

	pthread_key_create(&descriptor_page_pool_key, NULL);
	// allocating memory for the thread-local page_pool
	page_pool = __real_malloc(sizeof(descriptor_page_pool_t));

    pthread_setspecific(descriptor_page_pool_key, page_pool);

    return page_pool;
}

/*
 * scm_register_thread is called on a thread when it operates the first time
 * in libscm. The thread data structures are created or reused from previously
 * terminated threads.
 */
static stm_clock_t* scm_register_thread_clock(pthread_key_t* clock_key) {
    stm_clock_t* clock;
	descriptor_root_t* descriptor_root;

    lock_descriptor_roots();

//    if (init_threads == 0) {
        //initialize thread local data key and the global_time_lock spinlock
        //when the first thread is registered
        pthread_key_create(clock_key, NULL);
        pthread_spin_init(&global_time_lock, PTHREAD_PROCESS_PRIVATE);
//        init_threads = 1;
//    }

    if (terminated_descriptor_roots != NULL) {
        descriptor_root = terminated_descriptor_roots;
        terminated_descriptor_roots = terminated_descriptor_roots->next;
    } else {
        descriptor_root = new_descriptor_root();
    }
    scm_region_header_t* region = (scm_region_header_t*) init_region(sizeof(scm_region_header_t));
	region->dc = 0;

	clock = (stm_clock_t*) __real_malloc(sizeof(stm_clock_t));
	clock->descriptor_root = descriptor_root;
	clock->region = region;

    unlock_descriptor_roots();

    pthread_setspecific(*clock_key, clock);

    //assert: if descriptor_root belonged to a terminated thread,
    //block_thread was invoked on this thread
/*
 * FIXME:
 * scm_resume_thread_internal();
 * when this method is called, the descriptor_root pointer
 * is set to null.
 */

    return clock;
}

/*
 * scm_register_thread is called on a thread when it operates the first time
 * in libscm. The thread data structures are created or reused from previously
 * terminated threads.
 */
static descriptor_root_t *scm_register_thread() {
    descriptor_root_t *descriptor_root;

    lock_descriptor_roots();

    if (init_threads == 0) {
        //initialize thread local data key and the global_time_lock spinlock
        //when the first thread is registered
        pthread_key_create(&descriptor_root_key, NULL);
        pthread_spin_init(&global_time_lock, PTHREAD_PROCESS_PRIVATE);
        init_threads = 1;
    }

    if (terminated_descriptor_roots != NULL) {
        descriptor_root = terminated_descriptor_roots;
        terminated_descriptor_roots = terminated_descriptor_roots->next;
    } else {
        descriptor_root = new_descriptor_root();
    }

    unlock_descriptor_roots();

    pthread_setspecific(descriptor_root_key, descriptor_root);

    //assert: if descriptor_root belonged to a terminated thread,
    //block_thread was invoked on this thread
    scm_resume_thread_internal();

    return descriptor_root;
}

/*
 * scm_unregister_thread is called upon termination of a thread. The thread 
 * leaves the system and passes its data structures in a pool to be reused
 * by other threads upon creation.
 */
void scm_unregister_thread() {

    scm_block_thread_internal();

    descriptor_root_t *descriptor_root = get_descriptor_root();

    lock_descriptor_roots();

    descriptor_root->next = terminated_descriptor_roots;
    terminated_descriptor_roots = descriptor_root;

    unlock_descriptor_roots();
}

/*
 * scm_block_thread is called when a thread blocks to notify the system about it
 */
void scm_block_thread() {
    descriptor_root_t *descriptor_root = get_descriptor_root();

    //assert: we do not have the descriptor_roots lock
    lock_global_time();
    number_of_threads--;

    if (global_time == descriptor_root->global_phase) {
        //we have not ticked in this global period

        if (atomic_int_dec_and_test((int*) & ticked_threads_countdown)) {
            //we are the last thread to tick and therefore need to tick globally
            if (number_of_threads == 0) {
                ticked_threads_countdown = 1;
            } else {
                ticked_threads_countdown = number_of_threads;
            }
            global_time++;
        }
    }
    unlock_global_time();
}

/*
 * scm_resume_thread is called when a thread returns from blocking state to 
 * notify the system about it.
 */
void scm_resume_thread() {
    descriptor_root_t *descriptor_root = get_descriptor_root();

    //assert: we do not have the descriptor_roots lock
    lock_global_time();

    if (number_of_threads == 0) {
        /* if this is the first thread to resume/register,
         * then we have to tick to make
         * global progress, unless another thread registers
         * assert: ticked_threads_countdown == 1
         */
        descriptor_root->global_phase = global_time;
    } else {
        //else: we do not tick globally in the current global period
        //to avoid decrement of the ticked_threads_countdown
        descriptor_root->global_phase = global_time + 1;
    }
    number_of_threads++;

    unlock_global_time();
}

static descriptor_root_t *new_descriptor_root() {

    //allocate descriptor_root 0 initialized
    descriptor_root_t *descriptor_root =
            __real_calloc(1, sizeof (descriptor_root_t));

#ifdef SCM_PRINTOVERHEAD
    inc_overhead(__real_malloc_usable_size(descriptor_root));
#endif
#ifdef SCM_PRINTMEM
    printf("[new_descriptor_root (%u)]\n", __real_malloc_usable_size(descriptor_root));
    // TODO: uncomment inc_used_mem(__real_malloc_usable_size(descriptor_root));
#endif

    descriptor_root->globally_clocked_buffer.not_expired_length =
            SCM_MAX_EXPIRATION_EXTENSION + 3;
    descriptor_root->locally_clocked_buffer.not_expired_length =
            SCM_MAX_EXPIRATION_EXTENSION + 1;

    return descriptor_root;
}

static inline void lock_global_time() {
#ifdef SCM_PRINTLOCK
    if (pthread_spin_trylock(&global_time_lock)) {
        printf("thread %ld BLOCKS on global_time_lock\n", pthread_self());
        pthread_spin_lock(&global_time_lock);
    }
#else
    pthread_spin_lock(&global_time_lock);
#endif
}

static inline void unlock_global_time() {
    pthread_spin_unlock(&global_time_lock);
}

static inline void lock_descriptor_roots() {
#ifdef SCM_PRINTLOCK
    if (pthread_mutex_trylock(&terminated_descriptor_roots_mutex)) {
        printf("thread %ld BLOCKS on lock_descriptor_roots\n", pthread_self());
        pthread_mutex_lock(&terminated_descriptor_roots_mutex);
    }
#else
    pthread_mutex_lock(&terminated_descriptor_roots_mutex);
#endif
}

static inline void unlock_descriptor_roots() {
    pthread_mutex_unlock(&terminated_descriptor_roots_mutex);
}
